(*
* Copyright (C) 2003 Universitaet Koblenz-Landau
* 
* This file is part of KRHyper.
* 
* KRHyper is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
* 
* KRHyper is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
* details.
* 
* You should have received a copy of the GNU General Public License along with
* KRHyper; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA 02111-1307 USA
* 
* Contributor(s): Christoph Wernhard <wernhard@uni-koblenz.de>.
*)


type cfg = {
    mutable current_directory: string;
    mutable echo_flag: bool;
    mutable debug_flag: bool;
    mutable back_subsumption_flag: bool;
    mutable back_subsumption_within_news_flag: bool;
    mutable sort_heads_flag: bool;
    mutable check_disjunctions_are_ground_flag: bool;
    mutable proof_terms_flag: bool;
    mutable level_cut_flag: bool;
    mutable negative_units_flag: bool;
    mutable merge_inf_input_flag: bool;
    mutable wp_semantics_flag: bool;
    mutable show_derived_literals_flag: bool;
    mutable max_weight_initial_param: int;
    mutable max_weight_increment_param: int;
    mutable body_sorting_method_param: int;
    mutable weighing_method_param: int
  }

let get_current_directory cfg =
  cfg.current_directory

let set_current_directory cfg directory =
  cfg.current_directory <- directory

let get_echo_flag cfg = cfg.echo_flag
let set_echo_flag cfg bool = cfg.echo_flag <- bool
let get_debug_flag cfg = cfg.debug_flag
let set_debug_flag cfg bool = cfg.debug_flag <- bool
let get_back_subsumption_flag cfg = cfg.back_subsumption_flag
let set_back_subsumption_flag cfg bool =
  cfg.back_subsumption_flag <- bool
let get_back_subsumption_within_news_flag cfg = 
  cfg.back_subsumption_within_news_flag
let set_back_subsumption_within_news_flag cfg bool =
  cfg.back_subsumption_within_news_flag <- bool
let get_sort_heads_flag cfg = cfg.sort_heads_flag
let set_sort_heads_flag cfg bool = cfg.sort_heads_flag <- bool
let get_check_disjunctions_are_ground_flag cfg =
  cfg.check_disjunctions_are_ground_flag
let set_check_disjunctions_are_ground_flag cfg bool =
  cfg.check_disjunctions_are_ground_flag <- bool
let get_proof_terms_flag cfg = cfg.proof_terms_flag
let set_proof_terms_flag cfg bool = cfg.proof_terms_flag <- bool
let get_level_cut_flag cfg = cfg.level_cut_flag
let set_level_cut_flag cfg bool = cfg.level_cut_flag <- bool
let get_negative_units_flag cfg = cfg.negative_units_flag
let set_negative_units_flag cfg bool = cfg.negative_units_flag <- bool
let get_merge_inf_input_flag cfg = cfg.merge_inf_input_flag
let set_merge_inf_input_flag cfg bool = cfg.merge_inf_input_flag <- bool
let get_wp_semantics_flag cfg = cfg.wp_semantics_flag
let set_wp_semantics_flag cfg bool = cfg.wp_semantics_flag <- bool
let get_show_derived_literals_flag cfg = cfg.show_derived_literals_flag
let set_show_derived_literals_flag cfg bool =
  cfg.show_derived_literals_flag <- bool
let get_max_weight_initial_param cfg = cfg.max_weight_initial_param
let set_max_weight_initial_param cfg n = cfg.max_weight_initial_param <- n
let get_max_weight_increment_param cfg =
  cfg.max_weight_increment_param
let set_max_weight_increment_param cfg n =
  cfg.max_weight_increment_param <- n
let get_body_sorting_method_param cfg = cfg.body_sorting_method_param
let set_body_sorting_method_param cfg n = cfg.body_sorting_method_param <- n
let get_weighing_method_param cfg = cfg.weighing_method_param
let set_weighing_method_param cfg n = cfg.weighing_method_param <- n


let flags =
  [("back_subsumption",
    ("Use back subsumption - i.e. remove derived or input facts\n\
     if they are subsumed by a newly derived clause. [Works not with\n\
     disjunctions]",
       false,
       get_back_subsumption_flag, set_back_subsumption_flag));
   ("back_subsumption_within_news",
    ("Use back subsumption within the set of newly derived clauses.",
     false,
     get_back_subsumption_within_news_flag,
     set_back_subsumption_within_news_flag));
   ("sort_heads",
    ("Sort disjunctive heads heuristically.",
     true,
     get_sort_heads_flag, set_sort_heads_flag));
   ("check_disjunctions_are_ground",
    ("Verify at runtime for each derived disjunction that is it ground.",
     true,
     get_check_disjunctions_are_ground_flag,
     set_check_disjunctions_are_ground_flag));
   ("negative_units",
    ("Maintain negative units to detect refutations sooner and enable\n\
      complement splitting.",
      true,
     get_negative_units_flag, set_negative_units_flag));
   ("merge_inf_input",
    ("Merge facts inferred before any disjunctive branching into the input\n\
      facts. These derived facts then serve as lemmas in subsequent\n\
      tasks. This flag can be switched on and off around a call to 'run'\n\
      to generate certain lemmas.",
      false,
     get_merge_inf_input_flag, set_merge_inf_input_flag));
   ("wp_semantics",
    ("[Experimental] Enforce the `weak perfect' KRHyper semantics of\n\
      disjunctions. Overrides effects of other flags conflicting with\n\
      that semantics.",
     false,
     get_wp_semantics_flag, set_wp_semantics_flag));
   ("proof_terms",
    ("Generate proof terms for derived facts.",
     false,
     get_proof_terms_flag, set_proof_terms_flag));
   ("level_cut",
    ("Apply the level cut reduction. This is only effective if proof_terms\n\
      is also set.\n\
      Note: No level cut is applied at a node whose proof involves\n\
      stratified negation as failure or findall.",
     true,
     get_level_cut_flag, set_level_cut_flag));
   ("echo", 
    ("Print each input statement before it is processed.",
     false,
     get_echo_flag, set_echo_flag));
   ("debug",
    ("Debug mode. Print extra information about clause compilation.",
     false,
     get_debug_flag, set_debug_flag));
   ("show_derived_literals",
    ("Show facts as they are derived. May be useful for debugging.\n\
      Output are Prolog readable statements, one per line:\n\n  \
      derived_fact(Fact).  -- Fact has been derived in an extension step.\n  \
      attached_disjunction(DisjunctionId). -- The tableau has been\n      \
          extended  with a disjunction.\n  \
      selected_disjunct(Fact, DisjunctionId). -- Fact has been\n      \
       selected from an attached disjunction.\n\n\
     Notes: DisjunctionIds are currently not implemented.\n\
       Negative facts (see flag negative_units) are currently not considered.",
     false,
     get_show_derived_literals_flag, set_show_derived_literals_flag))]


let parameters =
  [("max_weight_initial", 
    ("The initial weight limit for inferred literals. If -1, weight\n\
      handling is not used. For infinite Herbrand domains, weight handling\n\
      can be required to ensure refutational completeness and\n\
      `finite-branch-model' completeness.",
     20,
     get_max_weight_initial_param, set_max_weight_initial_param));
   ("max_weight_increment", 
    ("The minimum amount by which the weight limit for inferred literals\n\
      is increased in a weight reset step. If max_weight_increment >= 0,\n\
      then the actual number used is the maximum of this number and the\n\
      smallest number by which an additional literal can possibly be\n\
      inferred. If max_weight_increment = -1, max_weight_initial is\n\
      used as weight limit and not incremented, yielding an obviously\n\
      incomplete calculus.",
      1,
      get_max_weight_increment_param,
      set_max_weight_increment_param));
   ("body_sorting_method", 
    ("Determines the heuristics for sorting body literals.\n\
      The following values can be used:\n\
      0 - A heuristics that ensures permissibility for builtins is used.\n\
      1 - The literal order of the input is used, except that delta\n    \
	  predicates are preferred. This may be useful on inputs\n    \
          resulting from program transformations.\n\
      2 - The literal order of the input is used.",
      0, 
      get_body_sorting_method_param,
      set_body_sorting_method_param));
   ("weighing_method", 
    ("Determines how the weight of a term is computed.\n\
      The following values can be used:\n\
      0 - The size of the term, i.e. the number of occurences of\n    \
	  predicate, function and constant symbols as well as variables\n    \
          in it is used.\n\
      1 - The depth of the term is used.",
     0,
     get_weighing_method_param,
     set_weighing_method_param))]
    
let set_to_defaults cfg =
  List.iter
    (fun (_, (_,default, _, writer)) ->
      (writer cfg default))
    flags;
  List.iter
    (fun (_, (_,default, _, writer)) ->
      (writer cfg default))
    parameters

let make_cfg () =
  let cfg =
    { current_directory=Sys.getcwd ();
      echo_flag=false;
      debug_flag=false;
      back_subsumption_flag=false;
      back_subsumption_within_news_flag=false;
      sort_heads_flag=false;
      check_disjunctions_are_ground_flag=false;
      negative_units_flag=false;
      merge_inf_input_flag=false;
      proof_terms_flag=false;
      level_cut_flag=false;
      wp_semantics_flag=false;
      show_derived_literals_flag=false;
      max_weight_initial_param=0;
      max_weight_increment_param=0;
      body_sorting_method_param=0;
      weighing_method_param=0
    } in
  set_to_defaults cfg;
  cfg
   
let print_smallheader x =
  print_string "----------------------------------------";
  print_string "--------------------------------------\n";
  print_string x;
  print_string "\n----------------------------------------";
  print_string "--------------------------------------\n"

let print_settings cfg =
  List.iter
    (fun (flag, (_, _, reader, _)) ->
      print_string "set_flag(";
      print_string flag;
      print_string ", ";
      print_string (if reader cfg then "true" else "false");
      print_endline ").")
    flags;
  List.iter
    (fun (flag, (_, _, reader, _)) ->
      print_string "set_parameter(";
      print_string flag;
      print_string ", ";
      print_int (reader cfg);
      print_endline ").")
    parameters

let print_flags_doc cfg =
  List.iter
    (fun (flag, (doc, default, _, _)) ->
      print_smallheader (Printf.sprintf "Flag: %s, default: %b" flag default);
      Printf.printf "%s\n\n" doc)
    flags;
  flush stdout

let print_parameters_doc cfg =
  List.iter
    (fun (param, (doc, default, _, _)) ->
      print_smallheader (Printf.sprintf "Parameter: %s, default: %d" 
			   param default);
      Printf.printf "%s\n\n" doc)
    parameters;
  flush stdout

let set_flag cfg flag newval =
  try
    let (_, _, _, writer) = List.assoc flag flags in
    writer cfg newval
  with Not_found ->
    raise (Err.Error "Unknown flag specified.")

let set_parameter cfg param newval =
  try
    let (_, _, _, writer) = List.assoc param parameters in
    writer cfg newval
  with Not_found ->
    raise (Err.Error "Unknown parameter specified.")


(****************************************************************************)

type verbosity = int

let current_verbosity = ref 3

let get_verbosity () = !current_verbosity

let set_verbosity n = current_verbosity := n

let current_msg_output = ref stderr

let get_msg_output () = !current_msg_output

let set_msg_output s =
  match s with
    "stdout" -> current_msg_output := stdout
  | _ -> current_msg_output := stderr
